/*
 * FILE:
 *   gfx_context.hpp
 *
 * PURPOSE:
 *   Interface for drawing onto a Window
 *
 * AUTHOR:
 *   Stephen Thompson <stephen@solarflare.org.uk>
 *
 * COPYRIGHT:
 *   Copyright (C) Stephen Thompson, 2008 - 2024.
 *
 *   This file is part of the "Coercri" software library. Usage of "Coercri"
 *   is permitted under the terms of the Boost Software License, Version 1.0, 
 *   the text of which is displayed below.
 *
 *   Boost Software License - Version 1.0 - August 17th, 2003
 *
 *   Permission is hereby granted, free of charge, to any person or organization
 *   obtaining a copy of the software and accompanying documentation covered by
 *   this license (the "Software") to use, reproduce, display, distribute,
 *   execute, and transmit the Software, and to prepare derivative works of the
 *   Software, and to permit third-parties to whom the Software is furnished to
 *   do so, all subject to the following:
 *
 *   The copyright notices in the Software and this entire statement, including
 *   the above license grant, this restriction and the following disclaimer,
 *   must be included in all copies of the Software, in whole or in part, and
 *   all derivative works of the Software, unless such copies or derivative
 *   works are solely in the form of machine-executable object code generated by
 *   a source language processor.
 *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 *   SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 *   FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *   DEALINGS IN THE SOFTWARE.
 *
 */

#ifndef COERCRI_GFX_CONTEXT_HPP
#define COERCRI_GFX_CONTEXT_HPP

#include "color.hpp"

#include "boost/shared_ptr.hpp"

#include <string>

namespace Coercri {

    class Font;
    class Graphic;
    class PixelArray;
    class Rectangle;
    class UTF8String;

    // This struct is used in plotPixelBatch.
    struct Pixel {
        Pixel(int x_, int y_, Color col_) : x(x_), y(y_), col(col_) { }
        
        int x;
        int y;
        Color col;
    };

    class GfxContext {
    public:
        // NOTE: If any "EndDrawing" type calls are required,
        // sub-classes should arrange for these to happen in the
        // destructor.
        virtual ~GfxContext() { }

        // Set/clear a clipping rectangle. This applies to all
        // graphics operations.
        virtual void setClipRectangle(const Rectangle &rect) = 0;
        virtual void clearClipRectangle() = 0;
        virtual Rectangle getClipRectangle() const = 0;

        // Get the total size of the target surface.
        virtual int getWidth() const = 0;
        virtual int getHeight() const = 0;
        
        // Drawing Functions
        virtual void clearScreen(Color col) = 0;   // (col.a is ignored)
        virtual void plotPixel(int x, int y, Color col) = 0;
        virtual void drawGraphic(int x, int y, const Graphic &graphic) = 0;
        virtual void drawGraphicModulated(int x, int y, const Graphic &graphic, Color col) = 0;

        // this just calls through to font.drawText.
        void drawText(int x, int y, const Font &font, const UTF8String &text, Color col);

        // The following have default implementations but can be
        // overridden if more efficient implementations are available.
        virtual void drawLine(int x1, int y1, int x2, int y2, Color col);
        virtual void drawRectangle(const Rectangle &rect, Color col);  // draws outline only.
        virtual void fillRectangle(const Rectangle &rect, Color col);  // draws a solid rectangle.
        virtual void plotPixelBatch(const Pixel *buf_start, int num_pixels);

        // Take a screenshot. Captures what has been drawn so far
        // into a PixelArray.
        // NOTE: This is a slow operation, it is mostly useful for testing.
        // NOTE: "alpha" values will be set to 255.
        virtual boost::shared_ptr<PixelArray> takeScreenshot() = 0;
    };

}

#endif
